/*
 * SubAtomHolder.cpp
 *
 * Copyright Notice: see copyright.txt
 *
 * Date: 3/8/2012 (D/M/Y)
 * Author: Dmitrij Lioubartsev
 * Email: dmitrijl42@gmail.com
 *
 * For class info, see Substitution.h
 * For algorithm description, see Substitution.h
 */

#include "Substitution.h"

/*
 * Initialize the Substitution class.
 * @param substitutes All atoms.
 * @param ratios A vector of pairs of (element, ratio). Sum of ratios should be 1.
 * @param excludes A vector of exclude strings.
 */
Substitution::Substitution(
		vector<Atom>* substitutes, vector<pair<string,double> >* ratios, vector<string>* excludes) {

	for(unsigned int i = 0; i < (*substitutes).size(); i++) {
		if((*substitutes)[i].numBonds() == 4)
			atoms.push_back(&((*substitutes)[i])); //only keep atoms with 4 bonds, because
													//thats what will be substituted
	}
	totalAtomCount = atoms.size();
	this->ratios = ratios;
	for(unsigned int i = 0; i < (*excludes).size(); i++) { //for each exclude string, create an exclusion
		(this->excludes).push_back(Exclusion((*excludes)[i]));
	}
}

/*
 * Wrapper method for the algorithm. Calls all algorithm functions
 * after each other and prints warning messages if something is wrong.
 * Returns true if everything went fine, otherwise returns false.
 */
bool Substitution::substitute() {
	bool ok = true;
	if(initSubsLeft()) { //step 1
		subsLeftcpy = subsLeft;
		setDistToSrcAll(); //step 2
		sort(atoms.begin(), atoms.end(), compDist); //step 3
		initPriorities(); //step 4
		if(subAll()) { //step 5, substitution itself
			shuffle(totalAtomCount*2); //step 6
		} else {
			cout << "WARNING: Could not substitute all atoms, by the given ratios." << endl
					<< " Advising to check configuration file. Elements/atoms not substituted: " << endl;
			cout << "Element | Number Atoms not substituted" << endl;
			map<string,int>::iterator it;
			for(it = subsLeft.begin(); it != subsLeft.end(); it++) {
				cout << it->first;
				int spaces = 10 - (it->first).length();
				for(int i = 0; i < spaces; i++) {
					cout.put(' ');
				}
				cout << it->second << endl;
			}
			ok = false;
		}
	} else if((*ratios).empty()){
		//do nothing
	} else {
		cout << "WARNING: total sum of given ratios is not 1, no"
				" substitution operations are executed." << endl;
		ok = false;
	}

	return ok;
}

Substitution::~Substitution() {
	// do nothing
}

/*
 * Step 1 (see Substitution.h)
 */
bool Substitution::initSubsLeft() {
	int sumAtoms = 0;
	double sumRatios = 0;
	for(unsigned int i = 0; i < (*ratios).size(); i++) {
		int tmp = ((double)totalAtomCount * (*ratios)[i].second); //the number of atoms of that type.
		subsLeft.insert(pair<string,int>((*ratios)[i].first,tmp)); //add to subsLeft
		sumAtoms += tmp;
		sumRatios += (*ratios)[i].second;
	}

	if(fabs(sumRatios-1) < 0.0001) { //if total ratios is 1
		//rounding errors may have lead to uneven number of atoms. Fix this error now.
		//example: total 100 atoms, ratios are si : 0.33, Al : 0.33, Br : 0.34
		//this will still lead to 33 atoms of each, even though all 100 atoms should be replaced.
		map<string,int>::iterator it = subsLeft.begin();
		while(sumAtoms != totalAtomCount) {
			if(sumAtoms < totalAtomCount) {
				(it->second)++;
				sumAtoms++;
			} else {
				(it->second)--;
				sumAtoms--;
			}
			it++;
			if(it == subsLeft.end())
				it = subsLeft.begin();
		}
		return true;
	} else {
		return false;
	}
}

/*
 * Step 2 (see Substitution.h)
 */
void Substitution::setDistToSrcAll() {
	queue<Atom*> atomsBFS;	//the queue for the Breadth-first search
	Atom* current = atoms[0]; //initialize current atom to first atom.

	/*
	 * For a general BFS algorithm see
	 * http://en.wikipedia.org/wiki/Breadth-first_search
	 */

	(*current).setDistToSrc(0); //initialize its distance to 0.
	vector<Atom*>* neighbors = (*current).getBonds(); //initialize neighbors
	for(unsigned int i = 0; i < (*neighbors).size(); i++) {
		(*((*neighbors)[i])).setDistToSrc(1); //set their distToSrc.
		atomsBFS.push((*neighbors)[i]); //add to queue
	}
	while(!atomsBFS.empty()) {
		current = atomsBFS.front();	//pop queue
		neighbors = (*current).getBonds();	//set neighbors to current's bonds
		for(unsigned int i = 0; i < (*neighbors).size(); i++) {
			if((*((*neighbors)[i])).getDistToSrc() == DEFAULT_DIST) { //if distance not yet set
				(*((*neighbors)[i])).setDistToSrc((*current).getDistToSrc() + 1); //set distance
				atomsBFS.push((*neighbors)[i]); //and push it to queue.
			}
		}
		atomsBFS.pop(); //current atom is treated. Pop it from queue.
	}

}

/*
 * Step 4 (see Substitution.h)
 */
void Substitution::initPriorities() {
	for(unsigned int i = 0; i < (*ratios).size(); i++) { // initialize priorities to 0
		priorities.push_back(pair<string,int>((*ratios)[i].first, 0));
	}

	for(unsigned int i = 0; i < excludes.size(); i++) {
		for(unsigned int j = 0; j < priorities.size(); j++) {
			if(excludes[i].getFirst() == priorities[j].first) //check first of Exclude
				(priorities[j].second) += 1;
			if(excludes[i].getSecond() == priorities[j].first) //check second of Exclude
				(priorities[j].second) += 1;
		}
	}
	//if first and second of the Exclude are equal, 2 will be added to its priority (eg. Al-O-Al)
 // now sort priorities, highest first

	sort(priorities.begin(),priorities.end(),prioComp);
}

/*
 * Step 5 (see Substitution.h)
 */
bool Substitution::subAll() {

	const int maxPrio = priorities[0].second;

	for(unsigned int i = 0; i < atoms.size(); i++) {
		Atom& a = *(atoms[i]); //current atom
		bool found = false;			//to indicate when sub is set
		int currentPrio = maxPrio;	//the current priority level to check
		vector<string> possibleSubs;//a vector of possible elements to sub with
		unsigned int prioInt = 0;	//index variable for priorities.
		while(!found) {
			if(prioInt == priorities.size()) //couldnt find anything to substitute with
				break;

			possibleSubs.clear(); //always start with clearing this
			while(priorities[prioInt].second == currentPrio && prioInt < priorities.size()) {
				//then, initialize possibleSubs to all elements of current priority
				possibleSubs.push_back(priorities[prioInt].first);
				prioInt++;
			}

			random_shuffle(possibleSubs.begin(), possibleSubs.end()); //to create some randomness.
			for(unsigned int j = 0; j < possibleSubs.size(); j++) {  //go through all possible subs
				if(legalSubstitute(a, possibleSubs[j])) { //if its ok to substitute
					a.setElement(possibleSubs[j]);		//substitute
					(subsLeft.at(possibleSubs[j]))--;	//update subsLeft
					found = true;						//and flag for finished.
					break;
				}
			}
			currentPrio = priorities[prioInt].second; //in case not found, go to next priority level.
		}
	}
	//now check if all was substituted ok.
	bool ok = true;
	map<string,int>::iterator it;
	for(it = subsLeft.begin(); it != subsLeft.end(); it++) {
		if(it->second != 0) {
			ok = false;
			break;
		}
	}

	return ok;
}

/*
 * Step 6 (see Substitution.h)
 * Picks atoms at random and tries to swap them, if it
 * does not violate any excludes. Do this 'seed' times.
 */
void Substitution::shuffle(int seed) {
	boost::random::mt19937 gen;
	gen.seed(time(0));
	boost::random::uniform_int_distribution<> dist(0, atoms.size()-1);
	for(int i = 0; i < seed; i++) {
		int int1 = dist(gen);
		int int2 = dist(gen);
		Atom& a1 = *(atoms[int1]);
		Atom& a2 = *(atoms[int2]);
		if(legalSubstitute(a1, a2.getElement(),false)
				|| legalSubstitute(a2, a1.getElement()),false) {
			string tmp = a1.getElement();
			a1.setElement(a2.getElement());
			a2.setElement(tmp);
		}
	}

}

/*
 * Returns true if substituting atom a with sub does not violate any
 * exclusions. If checkZero is true, a check will be done on subsLeft,
 * and it will return false if subsLeft is 0 for string "sub".
 */
bool Substitution::legalSubstitute(Atom& a, string sub, bool checkZero) {

	if(checkZero) {
		try{
			if(subsLeft.at(sub) == 0) {
				return false;
			}
		}
		catch(std::out_of_range&) {
			return false; //this should never happen, no matter the input.
		}

	}


	bool legal = true;
	string old(a.getElement()); //keep old atom element
	a.setElement(sub); //try substituting and then checking with all Exclusions.
	for(unsigned int i = 0; i < excludes.size(); i++) {
		if(a.checkPattern(excludes[i])) {
			legal = false; //done
			break;
		}
	}

	a.setElement(old); //put back whatever it had
	return legal;
}

void Substitution::printAtomsSubstituted() {
	map<string,int>::iterator it;
	cout << "Number of atoms substituted: " << endl;
	for(it = subsLeftcpy.begin(); it != subsLeftcpy.end(); it++) {
		cout << it->first << "       " << it->second << endl;
	}
}
